# -*- coding: utf-8 -*-
#!/usr/bin/env python

from functools import wraps, update_wrapper
from datetime import timedelta
import ConfigParser

import ujson as json
from redis import Redis, ConnectionError

from flask import make_response, request, current_app, Response, Flask, render_template, url_for

VERSION = "0.961"

config = ConfigParser.ConfigParser()
config.read("marlin.config")

if config.has_option("SERVER", "DEBUG"):
    DEBUG = config.get("SERVER", "DEBUG")
else:
    DEBUG = True
if config.has_option("REDIS", "REDIS_SERVER"):
    REDIS_SERVER = config.get("REDIS", "REDIS_SERVER")
else:
    REDIS_SERVER = "127.0.0.1"
if config.has_option("REDIS", "REDIS_PORT"):
    REDIS_PORT = config.get("REDIS", "REDIS_PORT")
else:
    REDIS_PORT = "6379"
if config.has_option("REDIS", "API_PRIFIX"):
    API_PREFIX = config.get("REDIS", "API_PREFIX")
else:
    API_PREFIX = '/api/'
if config.has_option("APP", "APP_NAME"):
    APP_NAME = config.get("APP", "APP_NAME")
else:
    APP_NAME = "marlin"
if config.has_option("REDIS", "REDIS_PASSWORD"):
    REDIS_PASSWORD = config.get("REDIS", "REDIS_PASSWORD")
else:
    REDIS_PASSWORD = ""

app = Flask(__name__)
app.debug = DEBUG


def cross_domain(origin=None, methods=None, headers=None,
                 max_age=21600, attach_to_all=True,
                 automatic_options=True):
    if methods is not None:
        methods = ', '.join(sorted(x.upper() for x in methods))
    if headers is not None and not isinstance(headers, basestring):
        headers = ', '.join(x.upper() for x in headers)
    if not isinstance(origin, basestring):
        origin = ', '.join(origin)
    if isinstance(max_age, timedelta):
        max_age = max_age.total_seconds()

    def get_methods():
        if methods is not None:
            return methods
        options_resp = current_app.make_default_options_response()
        return options_resp.headers['allow']

    def decorator(f):
        def wrapped_function(*args, **kwargs):
            if automatic_options and request.method == 'OPTIONS':
                resp = current_app.make_default_options_response()
            else:
                resp = make_response(f(*args, **kwargs))
            if not attach_to_all and request.method != 'OPTIONS':
                return resp
            h = resp.headers
            h['Access-Control-Allow-Origin'] = origin
            h['Access-Control-Allow-Methods'] = get_methods()
            h['Access-Control-Max-Age'] = str(max_age)
            if headers is not None:
                h['Access-Control-Allow-Headers'] = headers
            return resp

        f.provide_automatic_options = False
        return update_wrapper(wrapped_function, f)

    return decorator


def returns_json(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        r = f(*args, **kwargs)
        return Response(r, content_type='application/json; charset=utf-8')
    return decorated_function


@app.route("/")
def index():
    url_for('static', filename='toast.css')
    url_for('static', filename='marlin.jpg')
    return render_template('index.html')


def unified_router(rdm):
    error_response = Response(json.dumps(
        {'status': "Make sure redis-server is installed and running on http://%s:%s" % (REDIS_SERVER, REDIS_PORT)}),
                              content_type='application/json; charset=utf-8', status=500)
    if rdm:
        rdm.manipulate_data()
    else:
        return error_response
    if rdm.status:
        if rdm.data:
            return Response(rdm.string_data, content_type='application/json; charset=utf-8')
        elif rdm.method == "DELETE":
            return Response("", status=200, content_type='application/json; charset=utf-8')
        else:
            return Response(json.dumps({'status': "No data Found"}), content_type='application/json; charset=utf-8',
                            status=404)
    else:
        return error_response


@returns_json
@app.route(API_PREFIX + '<version>/<model>/<id>', methods=['GET', 'OPTIONS', 'POST', 'PUT', 'DELETE'])
@app.route(API_PREFIX + '<version>/<model>/<id>/', methods=['GET', 'OPTIONS', 'POST', 'PUT', 'DELETE'])
@cross_domain(origin='*')
def uni_api_router(version, model, id):
    rdm = RedisDatabaseManager(request, version, model, id)
    return unified_router(rdm)


@returns_json
@app.route(API_PREFIX + '<version>/<model>', methods=['GET', 'OPTIONS', 'POST', 'PUT', 'DELETE'])
@app.route(API_PREFIX + '<version>/<model>/', methods=['GET', 'OPTIONS', 'POST', 'PUT', 'DELETE'])
@cross_domain(origin='*')
def multi_api_router(version, model):
    rdm = RedisDatabaseManager(request, version, model, None)
    return unified_router(rdm)


@app.route("/ping/")
def ping():
    r = Redis(host=REDIS_SERVER, port=REDIS_PORT, password=REDIS_PASSWORD)
    try:
        r.ping()
        return Response("", status=200)
    except ConnectionError:
        return Response("", status=500)


class RedisDatabaseManager(object):
    """
    types:
        boolean
        float
        int
        string
        json

    NOT IMPLEMENTED
        foreignkey
        list
        manytomany

    if method == "POST" and not id:
        get_data()
    if method == "GET" and id:

    if method == "GET" and not id:

    if method == "DELETE" and id:

    if method == "DELETE" and not id:

    if method == "PUT" and id:
    """

    r = Redis(host=REDIS_SERVER, port=REDIS_PORT, password=REDIS_PASSWORD)
    app_name = APP_NAME

    def __init__(self, request, version, model, id):
        self.request = request
        self.version = version
        self.model = model
        self.id = id
        self.method = request.method
        self.key = "%s.%s.%s" % (self.app_name, self.version, self.model)
        self.status = False
        self.data = None
        self.string_data = ""
        self.length = 0

    def init_db(self):
        try:
            length = self.r.get(self.key + "_counter")
            if length:
                self.length = int(length)
        except ConnectionError:
            return False

    def manipulate_data(self):
        self.init_db()
        if self.method == "GET" and self.id:
            return self.get_from_redis()
        elif self.method == "GET" and not self.id:
            start = int(self.request.args.get("start", 1))
            end = int(self.request.args.get("end", self.length))
            return self.get_many_from_redis(start, end)
        elif (self.method == "POST" or self.method == "OPTIONS") and (not self.id or self.id == "add"):
            return self.set_to_redis()
        elif self.method == "DELETE" and self.id:
            self.delete_from_redis()
        elif self.method == "DELETE" and not self.id:
            if self.request.form.get("force") == "1":
                return self.flush_model()
            else:
                return self.delete_all_from_redis()
        elif (self.method == "PUT" or self.method == "POST") and self.id:
            return self.update_to_redis()

    def set_to_redis(self):
        try:
            kv_dict = {}
            for key in self.request.form.keys():
                try:
                    kv_dict[key] = float(self.request.form.get(key))
                    if int(kv_dict[key]) == kv_dict[key]:
                        kv_dict[key] = int(self.request.form.get(key))
                except ValueError:
                    kv_dict[key] = self.request.form.get(key)
            kv_dict['id'] = self.length + 1
            self.string_data = json.dumps(kv_dict)
            self.r.hset(self.key, self.length + 1, self.string_data)
            self.r.incr(self.key + "_counter")
            self.data = kv_dict
            self.status = True
        except ConnectionError:
            self.status = False

    def get_from_redis(self):
        try:
            self.data = self.r.hget(self.key, self.id)
            self.string_data = self.data
            self.status = True
        except ConnectionError:
            self.status = False

    def get_many_from_redis(self, start, end):
        try:
            id_list = range(start, end + 1)
            object_list = []
            if id_list:
                data_list = self.r.hmget(self.key, id_list)
                for item in data_list:
                    if item:
                        object_list.append(json.loads(item))
            self.data = object_list
            self.string_data = json.dumps(object_list)
            self.status = True
        except ConnectionError:
            self.status = False

    def delete_from_redis(self):
        try:
            self.r.hdel(self.key, self.id)
            self.status = True
        except ConnectionError:
            self.status = False

    def delete_all_from_redis(self):
        try:
            id_list = range(1, self.length + 1)
            for identity in id_list:
                self.r.hdel(self.key, identity)
            self.status = True
        except ConnectionError:
            self.status = False

    def update_to_redis(self):
        try:
            kv_dict = {}
            for key in self.request.form.keys():
                kv_dict[key] = self.request.form.get(key)
            kv_dict['id'] = self.id
            self.string_data = json.dumps(kv_dict)
            self.r.hset(self.key, self.id, self.string_data)
            self.data = kv_dict
            self.status = True
        except ConnectionError:
            self.status = False

    def flush_model(self):
        try:
            self.delete_all_from_redis()
            self.r.set(self.key + "_counter", 0)
            self.status = True
        except ConnectionError:
            self.status = False


import unittest
import marlin
import ujson as json


class MarlinTest(unittest.TestCase):

    def setUp(self):
        app = marlin.app
        app.config['TESTING'] = True
        self.client = app.test_client()

    def test_001_ping(self):
        resp = self.client.get("/ping/")
        self.assertTrue(resp.status, 200)

    def test_002_post_item(self):
        resp = self.client.post(path="/api/v1/fruits/", data={"name": "Apple", "calories": 52, "price": 120.00})
        self.assertTrue(json.loads(resp.data), dict(price=120.00, calories=52, name="Apple", id=1))

    def test_003_get_item(self):
        resp = self.client.get("/api/v1/fruits/1/")
        self.assertTrue(json.loads(resp.data), dict(price=120.00, calories=52, name="Apple", id=1))

    def test_004_get_list(self):
        resp = self.client.get("/api/v1/fruits/")
        self.assertTrue(json.loads(resp.data), list(dict(price=120.00, calories=52, name="Apple", id=1)))

    def test_005_delete_item(self):
        resp = self.client.delete("/api/v1/fruits/1/")
        self.assertTrue(resp.status, 200)

    def test_006_delete_all(self):
        resp = self.client.delete("/api/v1/fruits/")
        self.assertTrue(resp.status, 200)
        resp = self.client.post(path="/api/v1/fruits/", data={"name": "Apricot", "calories": 48, "price": 80.00})
        self.assertGreater(json.loads(resp.data).get("id"), 1)

    def test_007_flush_db(self):
        resp = self.client.delete("/api/v1/fruits/", data={'force': 1})
        self.assertTrue(resp.status, 200)
        resp = self.client.get("/api/v1/fruits/")
        self.assertTrue(json.loads(resp.data), {"status": "No data Found"})

if __name__ == '__main__':
    unittest.main()


# Marlin setup.py script
#
# It doesn't depend on setuptools, but if setuptools is available it'll use
# some of its features, like package dependencies.

from distutils.command.install_data import install_data
from distutils.command.install import INSTALL_SCHEMES
from glob import glob
import os
import sys
import marlin


class osx_install_data(install_data):
    # On MacOS, the platform-specific lib dir is /System/Library/Framework/Python/.../
    # which is wrong. Python 2.5 supplied with MacOS 10.5 has an Apple-specific fix
    # for this in distutils.command.install_data#306. It fixes install_lib but not
    # install_data, which is why we roll our own install_data class.

    def finalize_options(self):
        # By the time finalize_options is called, install.install_lib is set to the
        # fixed directory, so we set the installdir to install_lib. The
        # install_data class uses ('install_data', 'install_dir') instead.
        self.set_undefined_options('install', ('install_lib', 'install_dir'))
        install_data.finalize_options(self)


if sys.platform == "darwin":
    cmdclasses = {'install_data': osx_install_data}
else:
    cmdclasses = {'install_data': install_data}


def fullsplit(path, result=None):
    """
    Split a pathname into components (the opposite of os.path.join) in a
    platform-neutral way.
    """
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)

# Tell distutils to put the data_files in platform-specific installation
# locations. See here for an explanation:
# http://groups.google.com/group/comp.lang.python/browse_thread/thread/35ec7b2fed36eaec/2105ee4d9e8042cb
for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']

# Compile the list of packages available, because distutils doesn't have
# an easy way to do this.
packages, data_files = [], []
root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)


def is_not_module(filename):
    return os.path.splitext(filename)[1] not in ['.py', '.pyc', '.pyo']


for marlin_dir in ['marlin']:
    for dirpath, dirnames, filenames in os.walk(marlin_dir):
        # Ignore dirnames that start with '.'
        for i, dirname in enumerate(dirnames):
            if dirname.startswith('.'): del dirnames[i]
        if '__init__.py' in filenames:
            packages.append('.'.join(fullsplit(dirpath)))
            data = [f for f in filenames if is_not_module(f)]
            if data:
                data_files.append([dirpath, [os.path.join(dirpath, f) for f in data]])
        elif filenames:
            data_files.append([dirpath, [os.path.join(dirpath, f) for f in filenames]])


# Small hack for working with bdist_wininst.
# See http://mail.python.org/pipermail/distutils-sig/2004-August/004134.html
if len(sys.argv) > 1 and sys.argv[1] == 'bdist_wininst':
    for file_info in data_files:
        file_info[0] = '\\PURELIB\\%s' % file_info[0]

scripts = ['marlin/marlin-server']

version = marlin.marlin.VERSION

setup_args = {
    'name': 'marlin',
    'version': version,
    'url': 'http://atmb4u.github.io/marlin',
    'description': 'A fast and easy ReST API on top of redis',
    'author': 'Anoop Thomas Mathew',
    'author_email': 'atmb4u@gmail.com',
    'license': 'BSD',
    'packages': packages,
    'cmdclass': cmdclasses,
    'data_files': data_files,
    'scripts': scripts,
    'include_package_data': True,
    'classifiers': [
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',
    ]
}

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
else:
    setup_args['install_requires'] = [
        "redis>=2.9.1",
        "requests>=2.2.1",
        "ujson>=1.33"
    ]

setup(**setup_args)
